<!DOCTYPE html>
<html lang="en">
<body>
    <p>No pymbs configuration file was found at <span style="color:green"><b>{{ _pymbs_config_path }}</b></span></p>
    <p>A blank file has been located at that location.</p>
    <p>Please see <a href="{{ _help_url }}">{{ _help_url }}</a> for information on how to edit the file before proceeding.</p>
    <p><em>You will need to Restart the Kernel after editing the configuration file.</em></p>
</body>
</html>
